<li class="dropdown">
  <a href="{{ pathto(master_doc) }}"
     class="dropdown-toggle"
     data-toggle="dropdown">{{ _('Sections') }} <b class="caret"></b></a>
  <ul class="dropdown-menu globaltoc"
    >{{ toctree(maxdepth=theme_globaltoc_depth|toint, collapse=False) }}</ul>
</li>
