{%- if pagename != "search" %}
<form class="navbar-search pull-right" style="margin-bottom:-3px;" action="{{ pathto('search') }}" method="get">
  <input type="text" name="q" placeholder="Search" style="color:
  #333333; background-color: #dddddd;"/>
  <input type="hidden" name="check_keywords" value="yes" />
  <input type="hidden" name="area" value="default" />
</form>
{%- endif %}
