# ----------------------------------------------------------------
# SUNDIALS Copyright Start
# Copyright (c) 2002-2022, Lawrence Livermore National Security
# and Southern Methodist University.
# All rights reserved.
#
# See the top-level LICENSE and NOTICE files for details.
#
# SPDX-License-Identifier: BSD-3-Clause
# SUNDIALS Copyright End
# ----------------------------------------------------------------
sundials_version = 'v6.4.1'
arkode_version = 'v5.4.1'
cvode_version = 'v6.4.1'
cvodes_version = 'v6.4.1'
ida_version = 'v6.4.1'
idas_version = 'v5.4.1'
kinsol_version = 'v6.4.1'
year = '2022'
